define(['jquery','map','node','edge','mapView','geometryService', 'mapUtils', 'easeljs'], function($,Map,Node,Edge,MapView,GeometryService, MapUtils) {

    /**
     * Creates an instance of MapController
     *
     * @param {jQuery} $canvas
     * @param {string} mapId
     * @param {string} assetBasePath
     * @constructor
     */
    var MapController = function($canvas, mapId, assetBasePath)
    {
        if (!$canvas) throw "MapController: canvas is null";
        this.canvas = $canvas[0];
        this.id = mapId;
        this._assetBasePath = assetBasePath;
        this.nodeLen =0;
        this.currentNode="";
        this.stage = null;
        this.clear(); // it clears the current stage

        if (window.devicePixelRatio>1) {
        // grab the width and height from canvas
            var height = this.canvas.getAttribute('height');
            var width = this.canvas.getAttribute('width');
            // reset the canvas width and height with window.devicePixelRatio applied
            this.canvas.setAttribute('width', Math.round(width ));
            this.canvas.setAttribute('height', Math.round( height ));
            // force the canvas back to the original size using css
            this.canvas.style.width = (width/window.devicePixelRatio)+"px";
            this.canvas.style.height = (height/window.devicePixelRatio)+"px";
            // set CreateJS to render scaled
            //this.stage.scaleX = this.stage.scaleY = window.devicePixelRatio;
        }

        this.mapConnections = new Array();
        this.edgeCreated = [];

        /**
         * @name MapController#interactionMode
         *
         * Determines the interaction mode of the mapController, can be: [direct, hand]
         * - direct: default:interacts directly with shapes. (i.e drag & drop, resize,...)
         * - hand: click will move the map
         * @default direct
         * @type {string}
         */
        this.interactionMode=MapController.MODE.DIRECT;
        // Utility data
        this.dragStartPoint={x:0,y:0};
        this.dragging = false;
        this.dragStartPoint = {x:0,y:0};
        this.stageOffset = {x:0,y:0};
        this.connection = new createjs.Shape();

        /* used to calculate the extend of a map */
        this.mapExtend = {
            maxX: 0,
            maxY: 0
        };
        this.isSelectingEdge = false;
        this.isEditingEdge = false;
    };


    // CONSTANTS
    /**
     * Enum for interaction modes.
     *
     * @readonly
     * @enum {string}
     */
    MapController.MODE = {
        DIRECT: 'direct',
        HAND: 'hand'
    };

    Map.prototype.currentNode = "";

    // PUBLIC FUNCTIONS

    MapController.prototype.processMapJson = function(json)
    {
        if (json=="" || json==undefined || json == null){ /*??? servono tutte e 3 le condizioni? */
            MapController.map._imported = true;
        }
        this.setZoom(1);        // Uncaught TypeError: undefined is not a function :(
        this.createModel(json);
        this.createView(this.map, this.stage);
        document.app.useMapData();
    }

    /**
     * Loads a map.
     *
     */
    MapController.prototype.findMap = function()
    {
        this.clear();
        /*  if (this.id=='defaultMapId') return; non faccio il parsing del json */

        if (document.app.isDesktopDevice())
        {
            document.nativeService.findMapJson(this.id);
        }
        else
        {
            var json = document.nativeService.findMapJson(this.id);
            this.processMapJson(json);
        }
    };

    /**
     * Converts current map in Json.
     *
     * @returns {json}
     */
    MapController.prototype.toJson = function ()
    {
        if (!this.map) throw "MapController.toJson: map is null";
        return this.map.toJson();
    };

    /**
     * Clears stage, models and views.
     *
     */
    MapController.prototype.clear = function ()
    {
        // Deallocates widgets on stage
        if (this.stage)
        {
            // Remove stage children
            this.stage.removeAllChildren();
            // Remove html text items
            $('#node_contents').html('');
            $('#edge_contents').html('');
            $('#comment_contents').html('');
            // Remove stage handlers
            this.stage.off('stagemousedown');
            this.stage.off('stagemousemove');
            this.stage.off('stagemouseup');

        }
        // Create a new stage (detach old events)
        this.stage = new createjs.Stage(this.canvas);
        if (createjs.Touch.isSupported())
        {
            createjs.Touch.enable(this.stage);
        }
        // Reset the model (empty map)
        document.app.nodesCount = 0;
        this.createModel();
        // Reset the view (empty map)
        this.createView(this.map,this.stage);
    };

    /**
     * Redraws current map.
     *
     */
    MapController.prototype.redraw = function()
    {
        if (!this.stage) throw "MapController.redraw: stage is null";
        this.stage.update();
    };


    /**
     * Adds a node to current map
     * @param {json} jsonNode New node
     */

    MapController.prototype.addNode = function(jsonNode)
    {
        var me = this;
        // Create node model
        var node = this.map.addNode(jsonNode);
        //fix the dimension for dudat node with images
        // Create node view
        var nodeView = this.mapView.addNodeView(node, this.stage);
        this.calculateMapExtend();
        this.nodeLen++;
        // Assign node events
        nodeView.drawWidget();

        setTimeout(function(){
            document.app.onNodeStateChange(node);
            me.redraw();
        },100);
        this.setNodeHandlers(nodeView);
    };



    MapController.prototype.removeNodeView = function(node)
    {
        var nodeView = this.mapView.nodeViews[node.id];
        this.stage.removeChild(nodeView.widget);
        // delete all connected nodes
        this.deleteAllConnections(node.id);

        //si elimina anche il commento
        $('commentReadBtn.'+node.id).remove();
        delete this.mapView.nodeViews[node.id];
        this.redraw();


    };

    MapController.prototype.removeNode = function(node){
        delete this.map.nodes[node.id];
        this.nodeLen--;
    }

    MapController.prototype.updateNode = function(obj){
        var node = this.map.nodes[obj.id];
        node.loadJson(obj);
        var nodeView = this.mapView.nodeViews[node.id];
        nodeView.updateWidget();
        document.app.onNodeStateChange(node);
    }

    MapController.prototype.updateNodeForVariant = function(obj){
        var node = this.map.nodes[obj.id];
        node.loadJson(obj);
        var nodeView = this.mapView.nodeViews[node.id];
        nodeView.updateWidget();
        //document.app.onNodeStateChange(node);
    }


    MapController.prototype.removeEdgeView = function(edgeId){
        var edgeView = this.mapView.edgeViews[edgeId];
        if (edgeView) {
            this.deleteAConnection(edgeView.edge.id);
            delete this.mapView.edgeViews[edgeId];
        }
        this.redraw();
    }

    MapController.prototype.removeEdge = function(edgeId){
        //console.log(this);
        delete this.map.edges[edgeId];
        //this.nodeLen--;
    }

    MapController.prototype.deleteAConnection = function(edgeId)
    {
        var edgeView = this.mapView.edgeViews[edgeId];
        if (edgeView) {
            // some maps (maybe created with buggy apps?) have unconnected edges.
            if(edgeView.source != null && edgeView.target != null)
            {
                var edgeConnection = edgeView.source.id + edgeView.target.id;

                var index = this.mapConnections.indexOf(edgeConnection);
                if (index != -1) this.mapConnections.splice(index, 1);
            }
            this.stage.removeChild(edgeView.widget);
        }
        delete this.mapView.edgeViews[edgeId];
    }

    MapController.prototype.deleteAllConnections = function(nodeId)
    {
        for (var edgeId in this.map.edges)
        {
            var edge = this.map.edges[edgeId];
            if (edge.source==nodeId || edge.target==nodeId)
            {
                this.deleteAConnection(edgeId);
            }
        }
    }

    /**
     * Adds an edge to current map
     * @param {json} jsonEdge New edge
     */
    MapController.prototype.addEdge = function(jsonEdge)
    {
        // Create edge model

        var edge = this.map.addEdge(jsonEdge);

        this.edgeCreated.push(edge);
        // Create edge view
        var edgeView = this.mapView.addEdgeView(edge,this.map,this.stage);
        this.calculateEdgeIntersections(edgeView);
        this.calculateMapExtend();
        edgeView.updateWidget(this.map.nodes[edge.source],this.map.nodes[edge.target]);

        // Assign edge events
        this.setEdgeHandlers(edgeView);
    };

    MapController.prototype.addConnection = function(id)
    {
        this.mapConnections.push(id);
        //console.log(this.mapConnections.slice())
    }

    MapController.prototype.connectionExists = function(id)
    {
        //console.log("exists",this.mapConnections.indexOf(id));
        return this.mapConnections.indexOf(id)!=-1;
    }


    /**
     *
     * @returns {string}
     */
    MapController.prototype.getInteractionMode = function()
    {
        return this.interactionMode;
    };

    /**
     * Changes zoom factor
     * @param {number} zoom
     */
    MapController.prototype.setZoom = function(zoom)
    {
        this.stage.scaleX=zoom/100;
        this.stage.scaleY=zoom/100;

        zoom = zoom/100 * window.devicePixelRatio;

        this.stage.scaleX=zoom;
        this.stage.scaleY=zoom;
        if(document.app.node!=null){
            document.app.hideEditor();
            document.app.hideOverlayContents(document.app.node);
            document.app.hideCommentsBtn();
            document.app.placeCommentInAllNodes(this.id, this.stage.scaleX, this.stage.scaleY);
            document.app.fixCanvasSize();
        }

    };

    MapController.prototype.getZoom=function()
    {
        return {scaleX:this.stage.scaleX,scaleY:this.stage.scaleY};
    };

    MapController.prototype.getOffSet = function()
    {
        return {offSetX:this.stageOffset.x , offSetY:this.stageOffset.y};
    };

    MapController.prototype.getMapImages = function(){
        return this.map ? this.map.images : null;
    };


    // PRIVATE FUNCTIONS

    /**
     * @param {json} json Map json
     */
    MapController.prototype.createModel = function(json)
    {
        this.map = new Map(json, this._assetBasePath);

    };


    MapController.prototype.calculateMinExtend = function()
    {
        var minMapExtend = {minX: 0, minY: 0};
        for (var nodeId in this.mapView.nodeViews)
        {
            /* calculate the extend */
            var nodeView = this.mapView.nodeViews[nodeId];
            minXPos = nodeView.node.geometry.x - nodeView.node.geometry.width/2+100;
            minYPos = nodeView.node.geometry.y - nodeView.node.geometry.height/2+100;
            if (minMapExtend.minX>minXPos)
            {
                minMapExtend.minX = minXPos;
            }
            if (minMapExtend.minY>minXPos)
            {
                minMapExtend.minY = minXPos;

            }

        }

        for (var edgeId in this.mapView.edgeViews)
        {
            /* calculate the extend */

            var edgeView = this.mapView.edgeViews[edgeId];
            //var maxXPos = edgeView.area.width + edgeView.source.x;
            //var maxYPos = edgeView.area.height+ edgeView.source.y;
            var ed =  edgeView.edge;

            if (this.mapView.nodeViews[ed.source] == undefined)
                continue; // fix per eventuali errori di nodi che non sono linkati per via dell'undo redo
            var srcC = this.mapView.nodeViews[ed.source].node.geometry;
            var tgtC = this.mapView.nodeViews[ed.source].node.geometry;

            var edgeArea = MapUtils.prototype.normalizeArea(srcC.x, srcC.y, tgtC.x, tgtC.y);
            edgeArea = MapUtils.prototype.extendArea(edgeArea,ed.edgeData.ctrlPx,ed.edgeData.ctrlPy);

            minXPos = edgeArea.x1;
            minYPos =  edgeArea.y1;
            if (minMapExtend.minX>minXPos)
            {
                minMapExtend.minX = minXPos;
            }
            if (minMapExtend.minY>minXPos)
            {
                minMapExtend.minY = minXPos;

            }
        }

        return minMapExtend;

    };

    MapController.prototype.calculateMapExtend = function()
    {
        var mapExtend = {maxX: 0, maxY: 0};
        for (var nodeId in this.mapView.nodeViews)
        {
            /* calculate the extend */
            var nodeView = this.mapView.nodeViews[nodeId];
            maxXPos = nodeView.node.geometry.x + nodeView.node.geometry.width/2+100;
            maxYPos = nodeView.node.geometry.y + nodeView.node.geometry.height/2+100;
            if (mapExtend.maxX<maxXPos)
            {
                mapExtend.maxX = maxXPos;
            }
            if (mapExtend.maxY<maxYPos)
            {
                mapExtend.maxY = maxYPos;

            }

        }

        for (var edgeId in this.mapView.edgeViews)
        {
            /* calculate the extend */

            var edgeView = this.mapView.edgeViews[edgeId];
            //var maxXPos = edgeView.area.width + edgeView.source.x;
            //var maxYPos = edgeView.area.height+ edgeView.source.y;
            var ed =  edgeView.edge;

            if (this.mapView.nodeViews[ed.source] == undefined)
                continue; // fix per eventuali errori di nodi che non sono linkati per via dell'undo redo
            var srcC = this.mapView.nodeViews[ed.source].node.geometry;
            var tgtC = this.mapView.nodeViews[ed.source].node.geometry;

            var edgeArea = MapUtils.prototype.normalizeArea(srcC.x, srcC.y, tgtC.x, tgtC.y);
            edgeArea = MapUtils.prototype.extendArea(edgeArea,ed.edgeData.ctrlPx,ed.edgeData.ctrlPy);

            maxXPos = edgeArea.x2;
            maxYPos =  edgeArea.y2;
            if (mapExtend.maxX<maxXPos)
            {
                mapExtend.maxX = maxXPos;
            }
            if (mapExtend.maxY<maxYPos)
            {
                mapExtend.maxY = maxYPos;

            }
        }

        this.mapExtend = mapExtend;

        this.mapExtend.maxX = this.mapExtend.maxX * window.devicePixelRatio;
        this.mapExtend.maxY = this.mapExtend.maxY * window.devicePixelRatio;

        return this.mapExtend;
    }

    MapController.prototype.createView = function (map, stage)
    {
        if (!map) throw "MapController.createView: map is null";
        if (!stage) throw "MapController.createView: stage is null";
        var me = this;
        // Reference to current stage
        this.stage = stage;

        this.mapConnections = new Array();
        this.edgeCreated = [];


        // Create MapView elements (widgets) from model elements
        this.mapView = new MapView(stage, map);
        this.calculateMapExtend();

        /* calculate the new zoom */
        var newZoom = 1;
        if (this.mapExtend.maxX != 0){
            var scaleX =  this.canvas.width/this.mapExtend.maxX;
            scaleX = Math.min(1, scaleX);
            var scaleY =  this.canvas.height/this.mapExtend.maxY;
            //console.log(this.canvas.height);
            scaleY = Math.min(1, scaleY);
            newZoom = Math.min(scaleX, scaleY)*100;
        }else{
            newZoom = 100;
        }

        this.setZoom(newZoom);
        /* then it has to call a callback to set the zoom in the zoom controller */
        nativeApp.onjs_zoomChanged(newZoom);

        this.nodeLen = 0;

        var zoom =  this.getZoom();
        // Assign event handlers
        for (var nodeId in this.mapView.nodeViews)
        {

            this.nodeLen++;
            var nodeView = this.mapView.nodeViews[nodeId];
            this.setNodeHandlers(nodeView);
            nodeView.updateWidget();
            nodeView.updateWidget();
            document.app.placeCommentInNode(nodeView.node, zoom.scaleX, zoom.scaleY, this.getOffSet().offSetX, this.getOffSet().offSetY);
        }


        for (var edgeId in this.mapView.edgeViews)
        {
            var edgeView = this.mapView.edgeViews[edgeId];
            this.calculateEdgeIntersections(edgeView);
            edgeView.updateWidget(this.map.nodes[edgeView.edge.source],this.map.nodes[edgeView.edge.target]);
            var edgeMap = edgeView.edge.source+edgeView.edge.target;
            this.addConnection(edgeMap);
            this.setEdgeHandlers(edgeView);
        }

        this.redraw();

    };

    MapController.prototype.getNodeView = function(nodeId)
    {
        return this.mapView.nodeViews[nodeId];
    };

    MapController.prototype.getEdgeView = function(edgeId)
    {
        return this.mapView.edgeViews[edgeId];
    };

    MapController.prototype.updateMap = function(nodeId, pause) {
        console.log("STARTING UPDATE MAP");
        pause = pause || 0;
        var nodeView = this.mapView.nodeViews[nodeId],
            _this = this;

        nodeView.updateWidget();
        nodeView.drawWidget();

        this.updateEdge(nodeId);

        // so it automatically updates links
        this.updateEdgesLinkedTo(nodeId);

        setTimeout(function(){
            _this.redraw();
        }, pause);

        console.log("ENDING UPDATE MAP");
    };


    MapController.prototype.setNodeHandlers = function(nodeView)
    {
        var me = this;
        if (!nodeView) throw "MapController.setNodeHandlers: nodeView is null";
        // Assign high-level node events
        nodeView.onNodeDrag = function(evt) {me.nodeDragHandler(evt); };
        nodeView.onNodeEndDrag = function(evt) {me.nodeEndDragHandler(evt); };
        nodeView.onNodeDoubleClicked = function(evt){ me.onNodeDBLClick(evt); };
        nodeView.onModelChanged = function(evt,fromj,toj){ me.ModelChangedHandler(evt,fromj,toj); };
    };

    MapController.prototype.setEdgeHandlers = function(edgeView)
    {
        var me = this;
        // Assign edge events
        if (!edgeView) throw "MapController.setEdgeHandlers: edgeView is null";
        // Assign high-level edge events
        edgeView.onEdgeMidPointDrag = function(evt) {me.edgeMidPointDragHandler(evt); };
        edgeView.onEdgeMidPointEndDrag = function(evt) {me.edgeMidPointEndDragHandler(evt); };
        edgeView.onEdgeUpdated = function(evt) {me.onEdgeUpdatedHandler(evt); };
        edgeView.onModelChanged = function(evt,fromj,toj){ me.ModelChangedHandler(evt,fromj,toj); };
    };

    MapController.prototype.calculateEdgeIntersections = function(edgeView)
    {
        if (edgeView && edgeView.edge)
        {
            var sourceNodeView = this.mapView.nodeViews[edgeView.edge.source];
            var targetNodeView = this.mapView.nodeViews[edgeView.edge.target];
            /*
             commented because we want to enable users to choose a color for the edge which is independent from the source node
             if(sourceNodeView){
             //console.log(edgeView);
             //edgeView.edge.edgeData.penColor = sourceNodeView.node.nodeData.penColor;
             }*/

            GeometryService.prototype.calculateEdgeIntersections(edgeView,sourceNodeView,targetNodeView);
        }
    };

    MapController.prototype.switchInteractionModeToDirect = function() {
        this.stage.removeAllEventListeners();

        for (var nodeId in this.mapView.nodeViews)
        {
            var zoom =  this.getZoom();
            this.mapView.nodeViews[nodeId].on();

            document.app.placeCommentInNode(this.mapView.nodeViews[nodeId].node, zoom.scaleX, zoom.scaleY, this.getOffSet().offSetX, this.getOffSet().offSetY);

        }
        this.interactionMode = MapController.MODE.DIRECT;
    };

    MapController.prototype.switchInteractionModeToHand = function() {
        var me=this;

        document.app.hideEditor();
        document.app.hideCommentsBtn();
        document.app.hideOverlayContents(document.node);

        for (var nodeId in this.mapView.nodeViews)
        {
            this.mapView.nodeViews[nodeId].off();

        }

        this.stage.on('stagemousedown',function(evt)
        {
            me.dragging = true;
            me.dragStartPoint.x=evt.stageX;
            me.dragStartPoint.y=evt.stageY;

        });

        this.stage.on('stagemousemove',function(evt)
        {
            if (me.dragging)
            {
                var deltaX=evt.stageX-me.dragStartPoint.x;
                var deltaY=evt.stageY-me.dragStartPoint.y;
                var zoom =  me.getZoom();
                for (var nodeId in me.mapView.nodeViews)
                {
                    document.app.placeCommentInNode(me.mapView.nodeViews[nodeId].node, zoom.scaleX, zoom.scaleY, me.getOffSet().offSetX+deltaX, me.getOffSet().offSetY+deltaY);
                }

                me.stage.x=me.stageOffset.x+deltaX;
                me.stage.y=me.stageOffset.y+deltaY;
                me.redraw();
                document.app.updateEditor();
            }
        });

        this.stage.on('stagemouseup',function(evt)
        {
            me.dragging = false;
            me.stageOffset.x=me.stage.x;
            me.stageOffset.y=me.stage.y;
            var zoom =  me.getZoom();
            document.app.placeCommentInAllNodes(me.id, zoom.scaleX, zoom.scaleY);
            document.app.updateEditor();
        });

        this.stage.on('mouseleave',function(evt){
            me.dragging = false;
            me.stageOffset.x=me.stage.x;
            me.stageOffset.y=me.stage.y;
        });

        this.interactionMode = MapController.MODE.HAND;

    };


    MapController.prototype.updateEdge = function(nodeId)
    {


        for(var i=0;i<this.edgeCreated.length;i++){
            var edgeView = this.edgeCreated[i];
            var penColor = this.getNodeView(nodeId).node.nodeData.penColor;
            if(edgeView.source == nodeId)
            {

                //var e = this.mapView.edgeViews[edgeView.id];
                var e = this.getEdgeView(edgeView.id);

                if (e){
                    e.drawWidget();
                    e.updateWidget();
                }
                this.redraw();
            }
        }

    }

    MapController.prototype.updateEdgeFromJson = function(obj) {

        var edge = this.map.edges[obj.id]
        edge.loadJson(obj);
        var edgeView = this.mapView.edgeViews[edge.id];
        this.calculateEdgeIntersections(edgeView);
        this.updateEdgeProperties(edge.id);

        document.app.showLinkEditor(edge);
    }


    MapController.prototype.updateEdgeProperties = function(edgeId){
        var e = this.mapView.edgeViews[edgeId];
        if (!e) return;
        e.updateWidget(e.source, e.target);

        this.redraw();
    }


    MapController.prototype.updateEdgesLinkedTo = function(nodeId)
    {
        for (var edgeId in this.map.edges)
        {
            var edge = this.map.edges[edgeId];
            if (edge.source==nodeId || edge.target==nodeId)
            {
                var edgeView = this.mapView.edgeViews[edge.id];
                if (edgeView)
                {
                    this.calculateEdgeIntersections(edgeView);
                    edgeView.updateWidget(this.map.nodes[edge.source],this.map.nodes[edge.target]);
                }
            }
        }
    };


    MapController.prototype.ModelChangedHandler = function(evt,fromj,toj)
    {
    }


    MapController.prototype.nodeDragHandler = function(evt)
    {
        var node = this.map.nodes[evt.detail.id];
        $("."+node.id).hide();
        if (node)
        {
            document.app.node = node;
            // Update node model
            node.setCenter(evt.detail.x,evt.detail.y);
            // Update linked edges
            this.updateEdgesLinkedTo(evt.detail.id);
            document.app.hideEditor();
            document.app.hideOverlayContents(node);
            // Redraw
            this.redraw();
            var json = node.toJson();
            json.btbid = document.app.currentMapId;
            if(!this.fromJ) this.fromJ = JSON.stringify(json);
        }
    };

    MapController.prototype.nodeEndDragHandler = function(evt)
    {

        //document.app.fixCanvasSize();

        var node = this.map.nodes[evt.detail.id],
            commentReadBtn = $("."+node.id);

        if ( node.deepText.length > 0 ) {
            commentReadBtn.show();
        } else {
            commentReadBtn.hide();
        }
        if(this.fromJ)
        {
            var json = node.toJson();
            json.btbid = document.app.currentMapId;
            this.toJ = JSON.stringify(json);
            nativeApp.onjs_updateVariant(this.fromJ,this.toJ);
            delete this.fromJ;
            delete this.toJ;
        }
    };

    MapController.prototype.onNodeDBLClick = function(evt)
    {
        var node = this.map.nodes[evt.detail.id];
        var nodeView = this.mapView.nodeViews[node.id];
        this.currentNode = node;
        this.stage.setChildIndex(nodeView.widget,this.stage.getNumChildren() - 1);

        if (node) {
            document.app.onNodeStateChange(node);
        }

        this.redraw();
    };

    MapController.prototype.edgeMidPointDragHandler = function(evt)
    {
        var edge = this.map.edges[evt.detail.id];
        var edgeView = this.mapView.edgeViews[evt.detail.id];
        var source = this.map.nodes[edge.source];
        var target = this.map.nodes[edge.target];
        if (edge && edgeView && source && target)
        {
            var json = edge.toJson();
            json.btbid = document.app.currentMapId;
            if (!this.fromJ) this.fromJ = JSON.stringify(json);
            edge.edgeData.curve = true; // incurvo per fare le curve di bezier
            // Update edge model
            edge.setTextPoint(source.getCenter(),target.getCenter(), evt.detail.x,evt.detail.y);
            // Update edge View
            this.calculateEdgeIntersections(edgeView);
            edgeView.updateWidget(source,target);
            // Redraw
            this.redraw();


        }
    };


    MapController.prototype.edgeMidPointEndDragHandler = function(evt)
    {
        var edgeView = this.mapView.edgeViews[evt.detail.id];

        if(this.fromJ)
        {
            var json = edgeView.edge.toJson();
            json.btbid = document.app.currentMapId;
            console.log(this.fromJ);

            this.toJ = JSON.stringify(json);

            console.log('The object has changed, so we send onjs_updateVariant');
            console.log('FROM', this.fromJ);
            console.log('TO', this.toJ);
            nativeApp.onjs_updateVariant(this.fromJ,this.toJ);
            delete this.fromJ;
            delete this.toJ;
        }

    };

    MapController.prototype.onEdgeUpdatedHandler = function(evt)
    {
        //var edgeView = this.mapView.edgeViews[evt.detail.id];
        var edgeView = this.getEdgeView(evt.detail.id);

        if (edgeView)
        {
            this.calculateEdgeIntersections(edgeView);
            edgeView.updateWidget(this.map.nodes[edgeView.edge.source],this.map.nodes[edgeView.edge.target]);
            this.redraw();
            if(this.fromJ)
            {
                var json = edgeView.edge.toJson();
                json.btbid = document.app.currentMapId;
                this.toJ = JSON.stringify(json);
                console.log('The object has changed, so we send onjs_updateVariant');
                console.log('FROM', this.fromJ);
                console.log('TO', this.toJ);
                nativeApp.onjs_updateVariant(this.fromJ,this.toJ);
                delete this.fromJ;
                delete this.toJ;
            }
        }


    };


    MapController.prototype.setRoot = function(node,value)
    {
        // if the node has to be setted as root, we have to change mode of the other existing root node
        if (value==true){
            for (var nodeId in this.map.nodes) {

                this.map.nodes[nodeId].nodeData.mode = Node.MODE.CHILD;
                this.map.nodes[nodeId].setBrushColor(this.map.nodes[nodeId].nodeData.penColor, this.map.nodes[nodeId].nodeData.mode)
                //also labels are depending from colors
                this.mapView.nodeViews[node.id].setLabelPropForRootNode(this.map.nodes[node.id].nodeData.mode);

                var nodeView = this.mapView.nodeViews[nodeId];
                if(nodeView){
                    nodeView.updateWidget();
                    nodeView.drawWidget();
                }
            }
        }
        this.map.nodes[node.id].nodeData.mode = value === true ? Node.MODE.ROOT : Node.MODE.CHILD;

        this.mapView.nodeViews[node.id].setLabelPropForRootNode(this.map.nodes[node.id].nodeData.mode);
        this.redraw();
    };


    MapController.prototype.setEditableLabels = function(){
        for (var nodeId in this.mapView.nodeViews){

            this.map.nodes[nodeId].nodeData.textProperties.editing = false;
        }
    }

    MapController.prototype.startCreatingConnection = function(node)
    {
        node.isCreatingEdge = true;
    }


    MapController.prototype.createEdgeConnection = function(node, x, y)
    {
        var hasCreatedConnection = false;
        /* has to find if it is over a node */
        if (node.isCreatingEdge){
            var point = this.stage.globalToLocal(this.stage.mouseX, this.stage.mouseY);
            obj = this.stage.getObjectUnderPoint(point.x, point.y);
            for (var nodeId in this.mapView.nodeViews) {
                if (this.mapView.nodeViews[nodeId].widget==obj){
                    /* trovato il nodo, ora posso creare la connessione */
                    document.app.insertEdge(node.id, this.mapView.nodeViews[nodeId].node.id);
                    this.stage.setChildIndex( this.mapView.nodeViews[nodeId].widget,this.stage.getNumChildren() - 1);
                    this.stage.setChildIndex( this.mapView.nodeViews[node.id].widget,this.stage.getNumChildren() - 1);
                    this.redraw();
                    hasCreatedConnection = true;
                    break;
                }
            }
        }
        node.isCreatingEdge = false;
        return hasCreatedConnection;
    }

    MapController.prototype.createConnection = function()
    {
        //startingNode.isCreatingEdge = true;
    };

    MapController.prototype.drawConnection = function(x,y,x1,y1,node)
    {
        /*x = parseInt(x);
        y = parseInt(y);
        x1 = parseInt(x1);
        y1 = parseInt(y1);


        console.log(this.stage.globalToLocal(x, y));

        console.log('disegno width tratto -> colore', document.app.lastEdgeData.penWidth, document.app.lastEdgeData.penColor );
        // la linea e il colore
        this.connection.graphics
            .beginStroke(parseInt(document.app.lastEdgeData.penColor, 10))
        .setStrokeStyle(parseInt(document.app.lastEdgeData.penWidth, 10),"round")
        .moveTo(x, y)
        .lineTo(x1,y1);

        this.stage.addChild( this.connection );
        this.stage.update();*/
    };

    MapController.prototype.destroyConnection = function()
    {

    };


    MapController.prototype.setUpdateModelVariant = function(obj_from, obj_to){



        var json_from = obj_from.toJson();
        json_from.btbid = document.app.currentMapId;

        fromJ = JSON.stringify(json_from);

        var json_to = obj_to.toJson();
        json_to.btbid = document.app.currentMapId;
        toJ = JSON.stringify(json_to);

        if (fromJ != toJ)
        {
            console.log('The object has changed, so we send onjs_updateVariant');
            console.log('FROM', fromJ);
            console.log('TO', toJ);
            nativeApp.onjs_updateVariant(fromJ, toJ);
        }
    };


    return MapController;

});
